/****************************************************************
 * File: hm205.c
 *
 * Read Data from hm205-2 oscilloscope via vb-RS232 interface
 * RS-232-if: Siehe iX 1/95 S. 152 ff.
 *
 */
#define VERSION	"0.2"
/*
 ******************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <values.h>
#include <pwd.h>

#include <math.h>

/*------------------------- Constants ---------------------------*/
#define CONFIG_NAME			"hm205"
#define CONFIG_EXT			"conf"
#define CONFIG_ENV			"HM205"
#define PATH_DELIM			"/"
#define EXTENSION_DELIM			"."
#define ACT_DIR				"."

#ifndef DEFAULT_CONFIG_PATH
#  define DEFAULT_CONFIG_PATH		"/usr/local/lib"
#endif

#define COMMENT				"#"

#define CFG_COMMENT			COMMENT
#define CFG_DELIM			":"
#define CFG_X_SAMPLES_ID		"XSamplesPerDivision"
#define CFG_X_OFFSET_ID			"XOffset"
#define CFG_Y1_SAMPLES_ID		"Y1SamplesPerDivision"
#define CFG_Y1_OFFSET_ID		"Y1Offset"
#define CFG_Y2_SAMPLES_ID		"Y2SamplesPerDivision"
#define CFG_Y2_OFFSET_ID		"Y2Offset"
#define CFG_DEVICE_ID			"Device"
#define CFG_BAUD_ID			"Baudrate"


#define RTS_CTS_HSK
#undef 	XON_XOFF_HSK

#define	DEFAULT_DEV	"/dev/ttyS0"
#define DEFAULT_BAUD	38400

#ifndef LINE_BUFFER
#  define LINE_BUFFER		1024
#endif

#ifndef FLOAT_PRECISION
# define FLOAT_PRECISION	10
#endif

#define FLOAT_PREC_ENV_VAR_NAME 	"FLOAT_PRECISION"
static int FloatPrecision		= FLOAT_PRECISION;



#ifdef DEBUG
#  define DBG_PRINTF(Fmt, a...) {printf(Fmt, ## a);fflush(stdout);}
#else
#  define DBG_PRINTF(Fmt, a...)
#endif

/*------------------------- global variables --------------------*/
int SerFd 		= -1;
char *SerialName 	= DEFAULT_DEV;
char *CfgFileName 	= ACT_DIR PATH_DELIM CONFIG_NAME EXTENSION_DELIM CONFIG_EXT;
 
long 		BaudRate 	= DEFAULT_BAUD;
double 		XDotsPerDiv	= ((1023.0/10.0)*9.696969696969697/10.0);
double 		XDotsOffset	= 0.0;
double 		YDotsPerDiv[] 	= {
			0.0,							/* filler */ 
			((255.0/8.0)*(5.709803921568628-2.164705882352941)/4.0),/* Channel 1*/ 
			((255.0/8.0)*(5.772549019607843-2.164705882352941)/4.0),/* Channel 2*/
		};
				    
double 		YDotsOffset[]	= {
			0.0,				/* filler */
			(2.442477876106194-2.0),	/* Channel 1 */
			(2.400000000000000-2.0),		/* Channel 2 */
		};

double 		TperDiv		= 1.0;
double 		VperDiv		= 1.0;
double 		Voffset		= 0.0;
double		ProbeFactor	= 1.0;
int		Channel		= 0;

char Buffer[LINE_BUFFER];

int CfgLineCount = 0;

FILE *OutFile = NULL;

char *PrgName = NULL;

/*-------------------- Command line parameters --------------------------*/
#define PAR_TIME_BASE		't'
#define PAR_VOLTS_PER_DIV 	'y'
#define PAR_VOLTS_OFFSET	'Y'	
#define PAR_BAUD_RATE		'b'
#define PAR_SERIAL_LINE		's'
#define	PAR_CHANNEL		'c'
#define PAR_PROBE_FACTOR	'p'
#define NUM_PAR			7

static char CmdLine[] = {	PAR_TIME_BASE,		':',
  				PAR_VOLTS_PER_DIV,	':',
  				PAR_VOLTS_OFFSET,	':',
  				PAR_BAUD_RATE,		':',
  				PAR_SERIAL_LINE,	':',
  				PAR_CHANNEL,		':',
  				PAR_PROBE_FACTOR,	':',
  				/*--------------------------*/
  				'\0'				};


/*-------------------- HM205-Commands -----------------------------------*/
#define HMCMD_STATUS		's'
#define HMCMD_READ_DATA		'd'
#define HMCMD_RESET		'r'
#define HMCMD_WAIT_TE		't'
#define HMCMD_CH1_SEL		'1'
#define HMCMD_CH2_SEL		'2'
#define HMCMD_CH_UNSEL		'0'
#define HMCMD_RELEASE		'e'


#define HM_STATUS_REM		(1<<7)	/* not used */
#define HM_STATUS_RESET		(1<<6)	/* output */
#define HM_STATUS_TE		(1<<5)	/* input */
#define HM_STATUS_R_W		(1<<4)	/* not used */
#define HM_STATUS_SRQ		(1<<3)	/* output */
#define HM_STATUS_CLK_AZ	(1<<2)	/* output */
#define HM_STATUS_RES_AZ	(1<<1)	/* output */
#define HM_STATUS_CH		(1<<0)	/* output/input */

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
void Usage(void) {
  fprintf(stderr, "%s: Ver:" VERSION "(Built:" __DATE__ "/" __TIME__ ")\n",PrgName);
  fprintf(stderr, "Usage: %s -%c <ch> -%c <time> -%c <V/div> -%c <Y0>"
  		  " -%c <probe> -%c <baud> -%c <ser> <file>\n",		\
  	PrgName,	PAR_CHANNEL,		PAR_TIME_BASE,		\
  			PAR_VOLTS_PER_DIV,	PAR_VOLTS_OFFSET,	\
  			PAR_PROBE_FACTOR,	PAR_BAUD_RATE,		\
  			PAR_SERIAL_LINE					);  
  fprintf(stderr, "\t<ch>:    Channel 1, 2, 0: as selected at the oscilloscope\n");
  fprintf(stderr, "\t<time>:  Time per div in s (eg. 1e-3, 1m, 1ms = 1 ms/div)\n");
  fprintf(stderr, "\t<V/div>: Volts per div     (eg. 1e-3, 1m, 1mV = 1 mV/div)\n");
  fprintf(stderr, "\t<Y0>:    Y offset in div\n");
  fprintf(stderr, "\t<probe>: probe (eg. 10 = 1:10)\n");
  fprintf(stderr, "\t<baud>:  baudrate (default %d) \n",DEFAULT_BAUD);   
  fprintf(stderr, "\t<ser>:   serial device (default " DEFAULT_DEV "\n");
  fprintf(stderr, "\t<file>:  name of the output file\n");
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static double IntPow10(int Exponent) {
  int i; 
  double RetVal;

  RetVal = 1.0;
  if(Exponent > 0) for(i = 0; i <  Exponent; i++) RetVal *= 10.0;
  else	           for(i = 0; i >  Exponent; i--) RetVal /= 10.0;
  return(RetVal);
}


/*-------------------------------------------------------------------
 * 
 *---------------------------------------------------------------------------*/
static int GetIntPow10(double Value) {
  int i;

  i = 0;
  Value = fabs(Value);
  if(Value <= MINDOUBLE) return(0);
  if(Value >= 1.0) {
    while(Value >= 10.0) {
      i++;
      Value /= 10.0;
    }
    return(i);
  } else {
    while(Value < 1.0) {
      i++;
      Value *= 10.0;
    }
    return(-i);
  }
} 

/*-------------------------------------------------------------------
 * 
 *---------------------------------------------------------------------------*/
void PrintScientific(FILE *Output, double Value, char *Unit, int Sign) {
  int  ZehnerPotenz, Idx;
  char KiloMega[] 	= { 	' ',	/* 1 			*/ 
  				'k', 	/* 1e3 	= kilo 		*/
  				'M', 	/* 1e6 	= mega 		*/
  				'G', 	/* 1e9	= giga 		*/
  				'T', 	/* 1e12	= tera 		*/
  				'P',	/* 1e15	= peta 		*/
  				'E',	/* 1e18 = exa 		*/
  				'Z',	/* 1e21	= zetta 	*/
  				'Y',	/* 1e24	= yotta		*/
  				/*----------- ende -------------*/
  				'\0'};
  				
  char MiliMikro[] 	= { 	' ', 	/* 1			*/
  				'm', 	/* 1e-3	= milli		*/
  				'', 	/* 1e-6 = micro		*/
  				'n', 	/* 1e-9	= nano		*/
  				'p', 	/* 1e-12= pico		*/
  				'f', 	/* 1e-15= femto		*/
  				'a', 	/* 1e-18= atto		*/
  				'z',	/* 1e-21= zopto		*/
  				'y',	/* 1e-24= yocto		*/
  				/*----------- ende -------------*/	
  				'\0' };

  ZehnerPotenz = GetIntPow10(Value);
  if(Unit == NULL) Unit = "";
  if(ZehnerPotenz >= 0) {
    Idx = 0;
    while(Idx < ZehnerPotenz / 3) {
      if(KiloMega[Idx+1] == '\0') {
        fprintf(Output, "%e%s",Value, Unit);
        return;
      }
      else Idx++;
    }
    Value  /= IntPow10(Idx * 3);
    if(KiloMega[Idx] == ' ') { 
      if(Sign == 0) 	fprintf(Output, "%3.3f",Value);
      else 		fprintf(Output, "%+3.3f",Value);      
    } else { 
      if(Sign == 0) 	fprintf(Output, "%3.3f%c",Value, KiloMega[Idx]);
      else 		fprintf(Output, "%+3.3f%c",Value, KiloMega[Idx]);
    }
  } else {
    ZehnerPotenz = -ZehnerPotenz;
    Idx = 0;
    while(Idx <  ZehnerPotenz / 3) {
      if(MiliMikro[Idx+1] == '\0') {
        if(Sign == 0)	fprintf(Output, "%e%s",Value, Unit);
        else		fprintf(Output, "%+e%s",Value, Unit);
        return;
      }
      else Idx++;
    }
    if(ZehnerPotenz != (3 * Idx)) {
      if(MiliMikro[Idx+1] != '\0') { 
        Idx++;
      } else {
        if(Sign == 0)	fprintf(Output, "%e%s",Value, Unit);
        else		fprintf(Output, "%+e%s",Value, Unit);
        return;
      }
    }  
    Value *= IntPow10(Idx * 3);
    if(MiliMikro[Idx] == ' ') { 
      if(Sign == 0)	fprintf(Output, "%3.3f",Value);
      else		fprintf(Output, "%+3.3f",Value);
    } else { 
      if(Sign == 0) 	fprintf(Output, "%3.3f%c",Value, MiliMikro[Idx]);
      else		fprintf(Output, "%+3.3f%c",Value, MiliMikro[Idx]);
    }
  }
  fprintf(Output,"%s",Unit);
}

/*---------------------------------------------
 * Skip until the next white space (or end of line)
 *-----------------------------------------------*/
char *SkipPrint(char *String) {
  if(String == NULL) return(NULL);
  while((*String != '\0') && (*String != ' ') && (*String != '\t')) String++;   
  if(*String == '\0') return(NULL);
  return(String);  
}     

/*---------------------------------------------
 * Skip white spaces
 *-----------------------------------------------*/
char *SkipBlank(char *String) {
  if(String == NULL) return(NULL);
  while((*String != '\0') &&
       ((*String == ' ') || (*String == '\t') || 
        (*String == '\n')    || (*String == '\r')     )) { 
     String++;
  }
  if(*String == '\0') return(NULL);
  return(String);  
}

/*---------------------------------------------
 * Truncate trailing white spaces
 *-----------------------------------------------*/
void TruncBlank(char *String) {
  int i, Len;

  if(String == NULL) return;
  Len = strlen(String);
  String += Len;
  for(i = 0; i < Len; i++) {
    if((*String == ' ') || (*String == '\t') || (*String == '\n') || (*String == '\r')) 
      *String-- = '\0';
    else return;
  }
}
    
/*--------------------------------------------- 
 *
 *----------------------------------------------------------------------*/
char *GetLine(FILE *Fp) {
  char *Line;
  
  if(Fp == NULL) return(NULL);
  while((Line = fgets(Buffer, LINE_BUFFER, Fp)) != NULL) {
    CfgLineCount++;
    Line = SkipBlank(Line);
    if((Line != NULL) && (*Line != '\0')) {
      if(strncmp(Line, CFG_COMMENT, strlen(CFG_COMMENT)) != 0) return(Line);
    }
  }
  return(NULL);
}


/*-------------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
FILE *OpenConfigFile(void) {
  FILE *CfgFile;

  if((CfgFile = fopen(CfgFileName, "r")) != NULL) return(CfgFile);
  if(getenv(CONFIG_ENV) != NULL) {
    if((CfgFile = fopen(getenv(CONFIG_ENV), "r")) != NULL) return(CfgFile);
  }
  if((CfgFile = 				\
	fopen(DEFAULT_CONFIG_PATH PATH_DELIM CONFIG_NAME EXTENSION_DELIM CONFIG_EXT,"r")) \
	!= NULL) return(CfgFile);
  fprintf(stderr, "Warning: no config file found -- unsing default values\n");
  return(NULL);
}

/*-----------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
int CfgFloat(char *Line, double *Val) {
  Line = SkipBlank(Line);
  if(Line == NULL) return(1);
  if(strncmp(Line, CFG_DELIM, strlen(CFG_DELIM)) != 0) return(2);
  Line = SkipBlank(Line + strlen(CFG_DELIM));
  if(Line == NULL) return(3);
  *Val = atof(Line);
  return(0);
}

/*-----------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
int CfgInt(char *Line, long *Val) {
  Line = SkipBlank(Line);
  if(Line == NULL) return(1);
  if(strncmp(Line, CFG_DELIM, strlen(CFG_DELIM)) != 0) return(2);
  Line = SkipBlank(Line + strlen(CFG_DELIM));
  if(Line == NULL) return(3);
  *Val = atol(Line);
  return(0);
}

/*-----------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
int CfgString(char *Line, char *Val) {
  Line = SkipBlank(Line);
  if(Line == NULL) return(1);
  if(strncmp(Line, CFG_DELIM, strlen(CFG_DELIM)) != 0) return(2);
  Line = SkipBlank(Line + strlen(CFG_DELIM));
  if(Line == NULL) return(3);
  Val = Line;
  return(0);
}

/*-------------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
void GetConfigFile(void) {
  char *Line;  
  int ErrCode;
  FILE *Fp;
  char *SerDev = NULL;
  
  if((Fp = OpenConfigFile()) == NULL) return;
  while((Line = GetLine(Fp)) != NULL) {
    ErrCode = 0;
    if(strncasecmp(Line, CFG_X_SAMPLES_ID, strlen(CFG_X_SAMPLES_ID)) == 0) {
      ErrCode += CfgFloat(Line + strlen(CFG_X_SAMPLES_ID), &XDotsPerDiv); 
    } 
    if(strncasecmp(Line, CFG_X_OFFSET_ID, strlen(CFG_X_OFFSET_ID)) == 0) {
      ErrCode += CfgFloat(Line + strlen(CFG_X_OFFSET_ID), &XDotsOffset); 
    } 
    if(strncasecmp(Line, CFG_Y1_SAMPLES_ID, strlen(CFG_Y1_SAMPLES_ID)) == 0) {
      ErrCode += CfgFloat(Line + strlen(CFG_Y1_SAMPLES_ID), &(YDotsPerDiv[0])); 
    } 
    if(strncasecmp(Line, CFG_Y1_OFFSET_ID, strlen(CFG_Y1_OFFSET_ID)) == 0) {
      ErrCode += CfgFloat(Line + strlen(CFG_Y1_OFFSET_ID), &(YDotsOffset[0])); 
    } 
    if(strncasecmp(Line, CFG_Y2_SAMPLES_ID, strlen(CFG_Y2_SAMPLES_ID)) == 0) {
      ErrCode += CfgFloat(Line + strlen(CFG_Y2_SAMPLES_ID), &(YDotsPerDiv[1])); 
    } 
    if(strncasecmp(Line, CFG_Y2_OFFSET_ID, strlen(CFG_Y2_OFFSET_ID)) == 0) {
      ErrCode += CfgFloat(Line + strlen(CFG_Y2_OFFSET_ID), &(YDotsOffset[1])); 
    } 
    if(strncasecmp(Line, CFG_DEVICE_ID, strlen(CFG_DEVICE_ID)) == 0) {
      ErrCode += CfgString(Line + strlen(CFG_DEVICE_ID), SerDev);
      if(ErrCode != 0) {
        TruncBlank(SerDev);
        if((SerialName = strdup(SerDev)) == NULL) {
	  fprintf(stderr,"*** Malloc error ***\n");
	  exit(3);
        }
      } 
    } 
    if(strncasecmp(Line, CFG_BAUD_ID, strlen(CFG_BAUD_ID)) == 0) {
      ErrCode += CfgInt(Line + strlen(CFG_BAUD_ID), &BaudRate); 
    } 
    if(ErrCode != 0) {
      fprintf(stderr,"*** Error: Config File (line %d) is invalid ***\n", CfgLineCount);
      exit(3);
    }
  } /* while((Line = GetLine(CfgFile)) != NULL) */ 
}

/*--------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
double GetMultiplier(char *Mult) {

  if((Mult == NULL) || (*Mult == '\0')) return(1.0);
  switch(*Mult) {
    case 'Y': return(1e24);
    case 'Z': return(1e21);  
    case 'E': return(1e18);
    case 'P': return(1e15);    
    case 'T': return(1e12);  
    case 'G': return(1e9);
    case 'M': return(1e6);
    case 'k': return(1e3);  
    case 'm': return(1e-3);
    case 'u': return(1e-6);
    case '': return(1e-6);  
    case 'n': return(1e-9);  
    case 'p': return(1e-12);
    case 'f': return(1e-15);
    case 'a': return(1e-18);  
    case 'z': return(1e-21);
    case 'y': return(1e-24);
  } /* switch */
  return(1.0); 
}

/*--------------------------------------------------------------------- 
 *
 *-----------------------------------------------------------------------------*/

void GetCmdLine(int argc, char *argv[]) {
  int Cmd;
  char *Multiplier;


  do{  
    if((Cmd = getopt(argc, argv, CmdLine)) != EOF) {
      if((optarg != NULL) && (*optarg != '\0') && (*optarg == '=')) optarg++; 
      switch(Cmd) {
        case PAR_CHANNEL:
        	Channel = atoi(optarg);
          break;
        case PAR_TIME_BASE:
          	TperDiv = strtod(optarg, &Multiplier);
          	TperDiv *= GetMultiplier(Multiplier);
          break;
        case PAR_VOLTS_PER_DIV: 
          	VperDiv = strtod(optarg, &Multiplier);
		VperDiv *= GetMultiplier(Multiplier);
          break;
        case PAR_VOLTS_OFFSET:
          	Voffset = atof(optarg);
          break;
	case PAR_SERIAL_LINE:
	        TruncBlank(optarg);
		if((SerialName = strdup(optarg)) == NULL) {
		  fprintf(stderr,"*** Malloc error ***\n");
	 	  exit(3);
        	}
	  break;
	case PAR_BAUD_RATE:
		BaudRate = atol(optarg);
	  break;
	case PAR_PROBE_FACTOR:
		ProbeFactor = atof(optarg);
	  break;
	default:
	  fprintf(stderr,"*** error: invalid command line argumend '%c' (%d)***\n",
	  	Cmd, Cmd);
      }	/* switch() */
    } /* if((Cmd = getopt(a... */
  }while(Cmd != EOF);
  if(optind == argc) {
    Usage();
    exit(4);
  } else {
    DBG_PRINTF("Output: '%s'\n", argv[optind]);
    if((OutFile = fopen(argv[optind],"w")) == NULL) {
      fprintf(stderr,"*** Error: Cannot open output file '%s' ***\n", argv[optind]);
      exit(5); 
    }  
  }
} 


/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
unsigned long GetBaudFlag(long BaudRate) {
  switch(BaudRate) {
    case 50: 	return(B50);
    case 75: 	return(B75);
    case 110: 	return(B110);
    case 134: 	return(B134);
    case 150: 	return(B150);
    case 200: 	return(B200);
    case 300: 	return(B300);
    case 600: 	return(B600);
    case 1200: 	return(B1200);
    case 1800:	return(B1800);
    case 2400: 	return(B2400);
    case 4800: 	return(B4800);
    case 9600: 	return(B9600);
    case 19200: return(B19200);
    case 38400: return(B38400);
    case 57600: return(B57600);
    case 115200:return(B115200);
    case 230400:return(B230400);
    case 460800:return(B460800);
    default: return(0);
  } /* switch(BaudRate) */
  return(0);
} 

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
int CheckArguments(void) {
  int ErrCode = 0;
  
  if((Channel < 0) || (Channel > 2)) {
    fprintf(stderr, "Invalid Channel (%d) selected\n", Channel);
    ErrCode++;
  }
  if(TperDiv < 0.0) {
    fprintf(stderr,"Invalid Time/div (%e)\n", TperDiv);
    ErrCode++;
  }
  if(VperDiv < 0.0) {
    fprintf(stderr,"Invalid Volts/div (%e)\n", VperDiv);
    ErrCode++;
  }
  if(GetBaudFlag(BaudRate) == 0) {
    fprintf(stderr,"Invalid baudrate (%ld)\n", BaudRate);
    ErrCode++;
  }
  if(OutFile == NULL) {
    fprintf(stderr,"Invalid output file\n");
    ErrCode++; 	
  }
  if(SerFd < 0) {
    if((SerFd = open(SerialName, O_RDWR | O_NDELAY | O_SYNC)) < 0) {
       fprintf(stderr,"*** Error: Cannot open serial line '%s' ***\n", SerialName);
       ErrCode++;
    }
  }
  if((ProbeFactor < 1.0) || (ProbeFactor > 1e6)) {
    fprintf(stderr,"Invalid ProbeFactor 1:%e\n", ProbeFactor);
    ErrCode++;
  }  
  return(ErrCode);
} 

/*-----------------------------------------------
 *
 *-------------------------------------------------------------------------*/
void SendCmd(char Byte) {

  if(write(SerFd, &Byte, 1) != 1) {
    fprintf(stderr,"\n*** error, unable to send byte ***\n\n");
  }
}

/*------------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
int HexNibble2int(char Nibble) {

  if(Nibble <= '9')  	return(Nibble - '0');
  return(toupper(Nibble) - 'A' + 10);
}

/*-----------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
int ReadByte(void) {
  char c[4];
  if(read(SerFd, &c, 1) != 1) { 
    usleep(50);
    if(read(SerFd, &c, 1) != 1) return(-1);
  }
  return(c[0]);
}

/*-----------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
int GetHexDigit(void) {
  int c;
  
  c = ReadByte();
  if(c == '\n') return(0);
  if(!isxdigit(c)) return(-c);
  return(c);
}


/*-----------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
int ReadHexByte(void) {
  int c, d;

  if((c = GetHexDigit()) <= 0) return(-1);
  d = HexNibble2int(c) << 4;
  if((c = GetHexDigit()) <= 0) return(-2);
  d += HexNibble2int(c);
  return(d);
}

/*------------------------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
void FlushSerialLine() {
  while(ReadByte() >= 0);
} 

/*------------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
int GetStatus(void) {
  int c;

  FlushSerialLine();  
  SendCmd(HMCMD_STATUS);
  if(ReadByte() != HMCMD_STATUS) return(-1);
  if((c = ReadHexByte()) < 0) {
    fprintf(stderr, "*** Error: invalid Status (%d)***\n",c);
    return(-2);
  }
  DBG_PRINTF("Status: %d=0x%0X\n", c, c);
  DBG_PRINTF("\t ChI/ChII:   %d\n", c & HM_STATUS_CH 	 ? 1 : 0 );
  DBG_PRINTF("\t /ResAZ:     %d\n", c & HM_STATUS_RESET	 ? 1 : 0 );
  DBG_PRINTF("\t /ClkAZ:     %d\n", c & HM_STATUS_TE	 ? 1 : 0 );  
  DBG_PRINTF("\t Srq:        %d\n", c & HM_STATUS_R_W	 ? 1 : 0 );
  DBG_PRINTF("\t R/W (n.c.): %d\n", c & HM_STATUS_SRQ	 ? 1 : 0 );
  DBG_PRINTF("\t TE:         %d\n", c & HM_STATUS_CLK_AZ ? 1 : 0 );
  DBG_PRINTF("\t Reset:      %d\n", c & HM_STATUS_RES_AZ ? 1 : 0 );
  DBG_PRINTF("\t Rem (n.c.): %d\n", c & HM_STATUS_REM	 ? 1 : 0 );
  return(c);
}

/*---------------------------------------------------------------
 *
 *------------------------------------------------------------------------*/
int SelectChannel(int Ch) {

  FlushSerialLine();
  switch(Ch) {
    case 0:  
  	SendCmd(HMCMD_CH_UNSEL);
  	if(ReadByte() != HMCMD_CH_UNSEL) return(-1);
      break;
    case 1:
    	SendCmd(HMCMD_CH1_SEL);
    	if(ReadByte() != HMCMD_CH1_SEL) return(-2);
      break;
    case 2:
    	SendCmd(HMCMD_CH2_SEL);
    	if(ReadByte() != HMCMD_CH2_SEL) return(-3);
      break;
    default: return(-3);
  } /* switch(Ch) */
  return(0);
}

/*----------------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
int ReadData(void) {
  int Count = 0;
  int c;

  FlushSerialLine();
  SendCmd(HMCMD_READ_DATA);
  if(ReadByte() != HMCMD_READ_DATA) return(-1);
  fprintf(OutFile, COMMENT " Time [s]\t\tVolts \n");
  fprintf(OutFile, COMMENT "--------------------------------------------------------\n");
  while(1){
    if((c = ReadHexByte()) < 0) return(Count);
    fprintf(OutFile, "%1.*e\t%1.*e\n",	FloatPrecision, 	
		((double)Count/XDotsPerDiv-XDotsOffset)*TperDiv,
		FloatPrecision,
		((double)c/YDotsPerDiv[Channel]-YDotsOffset[Channel]-Voffset)*
			VperDiv*ProbeFactor					);
    Count++;
    DBG_PRINTF("%d: %d\n", Count, c);
  }; /*  while(1) */ 
  return(0);
}


/*-------------------------------------------------------
 *
 *-------------------------------------------------------------------------*/
#ifndef DEBUG
#define ShowPar()
#else
void ShowPar(void) {
  printf("Channel:	   %d\n",		Channel);
  printf("Timebase: 	   %e s/div\n", 	TperDiv);
  printf("Y:        	   %e V/div\n", 	VperDiv);
  printf("Y offset: 	   %e div\n",		Voffset);
  printf("Serial Line:	   %s@%ld Bd\n",	SerialName, BaudRate);
  printf("Fd:              %d\n",		SerFd);
  printf("XDotsPerDiv:	   %f\n",		XDotsPerDiv);
  printf("XDotsOffset:	   %f\n",		XDotsOffset);
  printf("YDotsPerDiv Ch1: %f\n",		YDotsPerDiv[0]);
  printf("YDotsOffset Ch1: %f\n",		YDotsOffset[0]);
  printf("YDotsPerDiv Ch2: %f\n",		YDotsPerDiv[1]);
  printf("YDotsOffset Ch2: %f\n",		YDotsOffset[1]);
}
#endif
/*----------------------------------------------------------------
 *
 *--------------------------------------------------------------------------*/
void ReleaseOszi(void) {

  FlushSerialLine();
  SendCmd(HMCMD_CH_UNSEL);
  FlushSerialLine();
  SendCmd(HMCMD_RELEASE);
}    

/*---------------------------------------------------------------------- 
 *
 *---------------------------------------------------------------------------*/
int main(int argc, char *argv[]) {
  struct termios t, Save_t;
  time_t Time;
  int Status;
  int ErrCode;
  char *EnvPtr;
  struct passwd *Pwd;
  
  if((PrgName = strrchr(argv[0],'/')) == NULL) PrgName = argv[0];
  else PrgName++;
  /*------------------- Message ---------------------------------------------*/
  printf("%s: Ver:" VERSION "(Built:" __DATE__ "/" __TIME__ ")\n",PrgName);
  printf("\tV.Bosch 2000\n");
  /*------------------- float precision -------------------------------------*/
  if((EnvPtr = getenv(FLOAT_PREC_ENV_VAR_NAME)) != NULL) {
    FloatPrecision = atoi(EnvPtr);
    if(FloatPrecision <= 0) FloatPrecision = FLOAT_PRECISION;
  }
  /*------------------- read config file ------------------------------------*/
  GetConfigFile();
  /*------------------- get command line arguments --------------------------*/
  GetCmdLine(argc, argv);
  /*------------------- check arguments -------------------------------------*/
  if(CheckArguments() != 0) {
    Usage();
    exit(3);
  }
  /*----------------------- init serial line --------------------------------*/
  tcgetattr(STDIN_FILENO, &t);
  Save_t = t;
  t.c_iflag = t.c_oflag = t.c_lflag = 0;
#if defined(RTS_CTS_HSK)
  /*------------- RTS/CTS-Handshake -----------------------------------------*/
  t.c_cflag = CREAD | CS8 | GetBaudFlag(BaudRate)  | HUPCL | CRTSCTS;  
#elif defined(XON_XOFF_HSK)
  /*------------- XON/XOFF-Handshake ----------------------------------------*/
  t.c_iflag = (IXON | IXOFF);
  t.c_cflag = CREAD | CS8 | GetBaudFlag(BaudRate)  | HUPCL;
#else 
  /*------------- No Handshake ----------------------------------------------*/
  t.c_cflag = CREAD | CS8 | GetBaudFlag(BaudRate)  | HUPCL ;   
#endif 

#ifdef CRLF
  t.c_oflag = (OPOST | ONLCR);          /* replace LF -> CRLF **/
#endif  
  t.c_cc[VMIN] 	= 0;
  t.c_cc[VTIME] = 0;  /* 0 Sek Time-Out */
  tcsetattr(SerFd, TCSANOW, &t);
  /*------------------ initialize oscilloscope ------------------------------*/
  SelectChannel(Channel);
  if((Status = GetStatus()) < 0) {
    fprintf(stderr,"*** Error: Oszi not connected to '%s' or switched off ***\n", SerialName);
    tcsetattr(SerFd, TCSANOW, &Save_t);
    close(SerFd);
    exit(1);
  }
  if((Status & HM_STATUS_TE) != 0) {
    fprintf(stderr,"*** Oszi ist nicht bereit: \n");
    fprintf(stderr,"\t-> Speicherbetrieb\n");
    fprintf(stderr,"\t-> Single Shot\n");
    ReleaseOszi();
  }
  if(Channel == 0) {
    Channel = ((Status & HM_STATUS_CH) == 0) ? 2 : 1;
    DBG_PRINTF("Channel %d selected\n", Channel);
  }
  ShowPar();
  /*----------------- Write file header -------------------------------------*/
  time(&Time);  
  fprintf(OutFile,COMMENT" Program: '%s', Ver."VERSION" (Built:"__DATE__"/"__TIME__")\n",
  		PrgName);
  fprintf(OutFile,COMMENT "\tV. Bosch 2000\n");
  fprintf(OutFile,COMMENT "\n");
  fprintf(OutFile,COMMENT "  File created %s", asctime(localtime(&Time)));
  Pwd = getpwuid((uid_t)geteuid());
  if(Pwd != NULL) {
     fprintf(OutFile,COMMENT "  by %s (uid: %s)\n", Pwd->pw_gecos, Pwd->pw_name);
  } /* if(Pwd != NULL) */
  fprintf(OutFile,COMMENT "------------------------------------------------------\n");
  fprintf(OutFile,COMMENT " Settings:\n");
  fprintf(OutFile,COMMENT "  Y-Range: %1.*e div : %1.*e div\n", 		\
  		FloatPrecision,	(0.0/YDotsPerDiv[Channel]-YDotsOffset[Channel]-Voffset)*VperDiv,	\
  		FloatPrecision, (255.0/YDotsPerDiv[Channel]-YDotsOffset[Channel]-Voffset)*VperDiv );
  if(Channel > 0) 
    fprintf(OutFile,COMMENT "  Oscilloscope channel %d\n", Channel); 
  fprintf(OutFile,COMMENT "  Time base:	");
  PrintScientific(OutFile, TperDiv, "s/div\n", 0);
  fprintf(OutFile,COMMENT "  Y-Gain:	");
  PrintScientific(OutFile, VperDiv, "V/div\n",  0);
  fprintf(OutFile,COMMENT "  Y-Offset:	%2.3f div\n",	Voffset);
  fprintf(OutFile,COMMENT "  Probe:	1:%3.2f\n",	ProbeFactor);
  fprintf(OutFile,COMMENT "\n");  
  /*----------------- read selected channel ---------------------------------*/ 
  if((ErrCode = ReadData()) < 0) {
    fprintf(stderr,"*** Fehler beim Auslesen ***\n");
  } 
  DBG_PRINTF("%d Datensaetze gelesen\n", ErrCode); 
  ReleaseOszi();
  tcsetattr(SerFd, TCSANOW, &Save_t);  
  close(SerFd);
  return(0);
} 
